///////////////////////////////////////////////////////////////////////////////
//
//  Copyright (c) 1999, Tres Seaver, Palladion Software. All rights reserved.
//  Orignal DUnit Copyright (C) 1999 Juancarlo Aez.
//  Original Java framework Copyright (C) 1999 by Kent Beck and Erich Gamma.
// 
//  LICENSE
//
//  Permission to reproduce and create derivative works from the
//  Software ("Software Derivative Works") is hereby granted to you
//  under the copyrights of Kent Beck, Erich Gamma, Juancarlo Aez,
//  and Tres Seaver (THE AUTHORS).  The authors also grant you the right
//  to distribute the Software and Software Derivative Works. 
//
//  The authors license the Software to you on an "AS IS" basis, without
//  warranty of any kind.  THE AUTHORS HEREBY EXPRESSLY DISCLAIMS ALL
//  WARRANTIES OR CONDITIONS, EITHER EXPRESS OR IMPLIED, INCLUDING, BUT
//  NOT LIMITED TO, THE IMPLIED WARRANTIES OR CONDITIONS OF
//  MERCHANTABILITY, NON INFRINGEMENT AND FITNESS FOR A PARTICULAR
//  PURPOSE.  You are solely responsible for determining the
//  appropriateness of using the Software and assume all risks
//  associated with the use and distribution of this Software, including
//  but not limited to the risks of program errors, damage to or loss of
//  data, programs or equipment, and unavailability or interruption of
//  operations.  THE AUTHORS WILL NOT BE LIABLE FOR ANY DIRECT DAMAGES
//  OR FOR ANY SPECIAL, INCIDENTAL, OR INDIRECT DAMAGES OR FOR ANY
//  ECONOMIC CONSEQUENTIAL DAMAGES (INCLUDING LOST PROFITS OR SAVINGS),
//  EVEN IF THE AUTHORS HAD BEEN ADVISED OF THE POSSIBILITY OF SUCH
//  DAMAGE.  The authors will not be liable for the loss of, or damage
//  to, your records or data, or any damages claimed by you based on a
//  third party claim.  
//
//  You agree to distribute the Software and any Software Derivatives
//  under a license agreement that: 1) is sufficient to notify all
//  licensees of the Software and Software Derivatives that the authors
//  assume no liability for any claim that may arise regarding the
//  Software or Software Derivatives, and 2) that disclaims all
//  warranties, both express and implied, from the authors regarding the
//  Software and Software Derivatives.  (If you include this Agreement
//  with any distribution of the Software and Software Derivatives you
//  will have meet this requirement).  You agree that you will not
//  delete any copyright notices in the Software. 
//
//  This Agreement is the exclusive statement of your rights in the
//  Software as provided by the authors.  Except for the licenses
//  granted to you in the second paragraph above, no other licenses are
//  granted hereunder, by estoppel, implication or otherwise.   
//
///////////////////////////////////////////////////////////////////////////////

unit DUnit_PEM;
//==============================================================================
//  Unit:       DUnit_PEM
//
//  Purpose:
//
//  Reimplement Juancarlo Aez' DUnit using a more "native" Delphi / component
//  approach (i.e., use Properties / Events / Methods, rather than
//  Interfaces / Inheritance).
//
//  Notes:
//
//  * This technique allows easy integration of the test framework into the
//    Delphi IDE -- test suites/case/results are simple non-visual components.
//
//  * Some variation along these lines might usefully backport to Java for
//    integration into an EJB environment (which uses essentially the same
//    P/E/M model, borrowed from Delphi).
//==============================================================================

interface

uses Classes        // TComponent, TList
   , SysUtils       // Exception, EAbort
   ;

type

  //============================================================================
  //  Forward declarations
  //============================================================================
  EAssertionFailed = class;
  TTestResult = class;
  TTestException = class;
  TTest = class;
  TTestCase = class;
  TTestSuite = class;

  //============================================================================
  //  Class:    EAssertionFailed
  //
  //  Responsibilities:
  //
  //  *  Thrown by test cases when an assertion fails.
  //============================================================================
  EAssertionFailed = class( EAbort )
	  constructor Create( msg : string = '');
  end;

  //============================================================================
  //  Class:    TTestResult
  //
  //  Responsibilities:
  //
  //  *  Collect test results across a series of test runs.
  //  *  Propagate changes to status of test run as events.
  //============================================================================
  TTestResultEvent = procedure( whichTest : TTest
                              ; res       : TTestResult ) of object;

  TTestResultExceptionEvent = procedure( whichTest : TTest
                                       ; res       : TTestResult
                                       ; e         : Exception ) of object;

  TTestResultListRange = 0..Classes.MaxListSize - 1;

  TTestResult = class( TComponent )

  private

    FOK               : Boolean;
    FTestsRunCount    : Integer;
    FErrors           : TList;
    FFailures         : TList;

    FOnTestStart      : TTestResultEvent;
    FOnTestError      : TTestResultExceptionEvent;
    FOnTestFailure    : TTestResultExceptionEvent;
    FOnTestEnd        : TTestResultEvent;

    procedure StartTest( whichTest  : TTest );
    procedure AddError( whichTest   : TTest
                      ; error       : Exception );
    procedure AddFailure( whichTest : TTest
                        ; failure   : Exception );
    procedure EndTest( whichTest    : TTest );

  protected

    function GetOK : Boolean;
    function GetTestsRunCount : Integer;

    function GetErrorCount : TTestResultListRange;
    function GetErrors( ndx : TTestResultListRange ) : TTestException;

    function GetFailureCount : TTestResultListRange;
    function GetFailures( ndx : TTestResultListRange ) : TTestException;

  public

    constructor Create( AOwner: TComponent );     override;
    destructor Destroy;                           override;

    property OK : Boolean
        read GetOK;

    procedure StopTesting;
    procedure RunTestCase( testcase : TTestCase );
    procedure Clear;

    property TestsRunCount : Integer
        read GetTestsRunCount;

    property ErrorCount : TTestResultListRange
        read GetErrorCount;

    property Errors[ ndx : TTestResultListRange ] : TTestException
        read GetErrors;

    property FailureCount : TTestResultListRange
        read GetFailureCount;

    property Failures[ ndx : TTestResultListRange ] : TTestException
        read GetFailures;

  published

    property OnTestStart : TTestResultEvent
        read FOnTestStart
       write FOnTestStart;

    property OnTestError : TTestResultExceptionEvent
        read FOnTestError
       write FOnTestError;

    property OnTestFailure : TTestResultExceptionEvent
        read FOnTestFailure
       write FOnTestFailure;

    property OnTestEnd : TTestResultEvent
        read FOnTestEnd
       write FOnTestEnd;
  end;


  //============================================================================
  //  Class:    TTestException
  //
  //  Responsibilities:
  //
  //  *  Capture data related to a failure (a failed assertion) or an error
  //     (some other exception) during a test run.
  //============================================================================
  TTestException = class

  private

    FAffectedTest       : TTest;
    FName               : String;
    FMessage            : String;

  public

    constructor Create( whichTest : TTest
                      ; e         : Exception
                      );

    property AffectedTest : TTest
        read FAffectedTest;

    property Name : String
        read FName;

    property Message : String
        read FMessage;
  end;


  //============================================================================
  //  Class:    TTest
  //
  //  Responsibilities:
  //
  //  *  Abstract base for runnable tests.
  //  *  Know parent (enclosing test suite), if any.
  //  *  Know test result,  if any.
  //============================================================================
  TTest = class( TComponent )

  private

    FParent         : TTestSuite;
    FTestResult     : TTestResult;

  protected

    procedure SetParent( newParent : TTestSuite );
    procedure SetTestResult( newResult : TTestResult );     virtual;

    procedure Notification( AComponent  : TComponent
                          ; Operation   : TOperation );     override;


  public

    constructor Create( AOwner : TComponent );              override;
    destructor Destroy;                                     override;

    procedure Run;                                          virtual; abstract;

  published

    property Parent : TTestSuite
        read FParent
       write SetParent;

    property TestResult : TTestResult
        read FTestResult
       write SetTestResult;
  end;


  //============================================================================
  //  Class:    TTestCase
  //
  //  Responsibilities:
  //
  //  *  Simple, "leaf" test.
  //  *  Provide facilities for external code to assert conditions against
  //     tested state.
  //  *  Allow clients to customize preparation/execution/cleanup of test
  //     via event handlers.
  //============================================================================
  TTestCaseEvent = procedure( testCase : TTestCase ) of object;

  TTestCase = class( TTest )

  private

    FOnSetup    : TTestCaseEvent;
    FOnExecute  : TTestCaseEvent;
    FOnTeardown : TTestCaseEvent;

    procedure Fail( what    : String
                  ; fmt     : String
                  ; args    : array of const
                  );

  protected

    procedure Setup;
    procedure DoRunTest;
    procedure Teardown;

  public

    constructor Create( AOwner : TComponent );              override;
    destructor Destroy;                                     override;

    procedure Run;                                          override;
    procedure RunDirectTest;

    procedure Assert( condition : Boolean
                    ; msg       : String ='' );

    procedure AssertAssigned( obj   : TObject 
                            ; msg   : String = ''
                            );

    procedure AssertUnassigned( obj   : TObject 
                              ; msg   : String = ''
                              );

    procedure AssertIdentity( expected  : TObject 
                            ; actual    : TObject
                            ; msg       : String = ''
                            );

    procedure AssertLongValue( expected : LongInt
                             ; actual   : LongInt
                             ; msg      : String ='' 
                             );

    procedure AssertDoubleValue( expected : Double
                               ; actual   : Double
                               ; delta    : Double
                               ; msg      : String = ''
                               );

  published

    property OnSetup : TTestCaseEvent
        read FOnSetup
       write FOnSetup;

    property OnExecute : TTestCaseEvent
        read FOnExecute
       write FOnExecute;

    property OnTeardown : TTestCaseEvent
        read FOnTeardown
       write FOnTeardown;
  end;


  //============================================================================
  //  Class:    TTestSuite
  //
  //  Responsibilities:
  //
  //  *  Composite test (manage list of contained tests).
  //  *  Execute by iteration over contained tests.
  //  *  Propagate notice of batch start/stop via events.
  //============================================================================
  TTestSuiteEvent = procedure( testSuite : TTestSuite ) of object;

  TTestSuite = class( TTest )

  private

    FTests      : TList;
    FOnBeginRun : TTestSuiteEvent;
    FOnEndRun   : TTestSuiteEvent;

  protected

    procedure SetTestResult( newResult : TTestResult );     override;

    procedure Notification( AComponent  : TComponent
                          ; Operation   : TOperation );     override;

  public

    constructor Create( AOwner : TComponent );              override;
    destructor Destroy;                                     override;

    procedure AddTest( child : TTest );
    procedure RemoveTest( child : TTest );

    procedure Run;                                          override;

  published

    property OnBeginRun : TTestSuiteEvent
        read FOnBeginRun
       write FOnBeginRun;

    property OnEndRun : TTestSuiteEvent
        read FOnEndRun
       write FOnEndRun;
  end;

//==============================================================================
//  Function:   Register
//
//  Purpose:    Install TTestResult, TTestCase, and TTestSuite into the VCL.
//==============================================================================
procedure Register;

implementation

//==============================================================================
//  EAssertionFailed implementation
//==============================================================================

constructor EAssertionFailed.Create( msg : string = '');
begin
    inherited Create( msg );
end;


//==============================================================================
// TTestResult implementation
//==============================================================================

constructor TTestResult.Create( AOwner : TComponent );
begin
    inherited Create( AOwner );
    FOK                 := True;
    FTestsRunCount      := 0;
    FErrors             := TList.Create;
    FFailures           := TList.Create;
end;

destructor TTestResult.Destroy;
begin
    FFailures.Free;
    FErrors.Free;
    inherited Destroy;
end;

procedure TTestResult.StopTesting;
begin
    FOK                 := False;
end;

procedure TTestResult.StartTest( whichTest : TTest );
begin
    Inc( FTestsRunCount );
    if Assigned( FOnTestStart ) then
        FOnTestStart( whichTest, self );
end;


procedure TTestResult.RunTestCase( testcase : TTestCase );
begin
    self.StartTest( testcase );
    try

      try
          testcase.RunDirectTest();

      except

         on e : EAssertionFailed do
            self.AddFailure( testcase, e );

         on e : Exception do
            self.AddError( testcase, e );
      end;

    finally
      self.EndTest( testcase );
    end
end;

procedure TTestResult.Clear;
var
    iError      : Integer;
    iFailure    : Integer;
begin
    FTestsRunCount := 0;
    for iError := 0 to ErrorCount - 1 do
        TObject( FErrors[ iError ] ).Free;
    FErrors.Clear;
    for iFailure := 0 to FailureCount - 1 do
        TObject( FFailures[ iFailure ] ).Free;
    FFailures.Clear;
end;

procedure TTestResult.AddError( whichTest : TTest
                              ; error     : Exception );
begin
    FErrors.Add( TTestException.Create( whichTest, error ) );
    if Assigned( FOnTestError ) then
        FOnTestError( whichTest, self, error );
end;

procedure TTestResult.AddFailure( whichTest : TTest
                                ; failure   : Exception );
begin
    FFailures.Add( TTestException.Create( whichTest, failure ) );
    if Assigned( FOnTestFailure ) then
        FOnTestFailure( whichTest, self, failure );
end;

procedure TTestResult.EndTest( whichTest : TTest );
begin
    if Assigned( FOnTestEnd ) then
        FOnTestEnd( whichTest, self );
end;

function TTestResult.GetOK : Boolean;
begin
    Result := FOK;
end;

function TTestResult.GetTestsRunCount : Integer;
begin
    Result := FTestsRunCount;
end;

function TTestResult.GetErrorCount : TTestResultListRange;
begin
    Result := FErrors.Count;
end;

function TTestResult.GetErrors( ndx : TTestResultListRange ) : TTestException;
begin
    Result := FErrors[ ndx ];
end;

function TTestResult.GetFailureCount : TTestResultListRange;
begin
    Result := FFailures.Count;
end;

function TTestResult.GetFailures( ndx : TTestResultListRange ) : TTestException;
begin
    Result := FFailures[ ndx ];
end;


//==============================================================================
//  TTestException implementation
//==============================================================================

constructor TTestException.Create( whichTest : TTest
                                 ; e         : Exception
                                 );
begin
    FAffectedTest   := whichTest;
    FName           := e.ClassName;
    FMessage        := e.Message;
end;


//==============================================================================
//  TTest implementation
//==============================================================================

constructor TTest.Create( AOwner : TComponent );
begin
    inherited Create( AOwner );
end;

destructor TTest.Destroy;
begin
    inherited Destroy;
end;

procedure TTest.SetParent( newParent : TTestSuite );
begin
    if Assigned( FParent ) then
        FParent.RemoveTest( self );
    FParent := newParent;
    if Assigned( FParent ) then
    begin
      FParent.AddTest( self );
      if not Assigned( TestResult ) then
         SetTestResult( FParent.TestResult );
    end;
end;

procedure TTest.SetTestResult( newResult : TTestResult );
begin
    FTestResult := newResult;
end;

procedure TTest.Notification( AComponent   : TComponent
                            ; Operation    : TOperation );
begin
    if ( AComponent is TTest )
       and ( Operation = opRemove )
       and ( AComponent = FParent ) then
        FParent := Nil;
end;


//==============================================================================
//  TTestCase implementation
//==============================================================================

constructor TTestCase.Create( AOwner : TComponent );
begin
    inherited Create( AOwner );
end;

destructor TTestCase.Destroy;
begin
    inherited Destroy;
end;

procedure TTestCase.Fail( what  : String
                        ; fmt   : String
                        ; args  : array of const
                        );
var
    realFmt : String;
begin
    realFmt := Format( '[%s:%s] %s', [ Name, what, fmt ] );
    raise EAssertionFailed.CreateFmt( realFmt, args );
end;

procedure TTestCase.Assert( condition : Boolean
                          ; msg       : String ='' );
begin
    if not condition then
        Fail( 'Assert', '%s', [ msg ] );
end;

procedure TTestCase.AssertAssigned( obj   : TObject 
                                  ; msg   : String = ''
                                  );
begin
    if not Assigned( obj ) then
        Fail( 'AssertAssigned', '%s', [ msg ] );
end;

procedure TTestCase.AssertUnassigned( obj   : TObject 
                                    ; msg   : String = ''
                                    );
begin
    if Assigned( obj ) then
        Fail( 'AssertUnassigned', '%s', [ msg ] );
end;


procedure TTestCase.AssertIdentity( expected  : TObject 
                                  ; actual    : TObject
                                  ; msg       : String = ''
                                  );
begin
    if not ( expected = actual ) then
        Fail( 'AssertIdentity', '%s', [ msg ] );
end;

procedure TTestCase.AssertLongValue( expected : LongInt
                                   ; actual   : LongInt
                                   ; msg      : String ='' 
                                   );
begin
    if not ( expected = actual ) then
        Fail( 'AssertLongValue', '%s -- expected %d, actual %d'
            , [ msg, expected, actual ] );
end;

procedure TTestCase.AssertDoubleValue( expected : Double
                                     ; actual   : Double
                                     ; delta    : Double
                                     ; msg      : String = ''
                                     );
begin
    if not ( Abs( expected - actual ) <= delta ) then
        Fail( 'AssertDoubleValue', '%s -- expected %g, actual %g'
            , [ msg, expected, actual ] );
end;

procedure TTestCase.Setup;
begin
    if Assigned( FOnSetup ) then FOnSetup( self );
end;

procedure TTestCase.DoRunTest;
begin
    if Assigned( FOnExecute ) then FOnExecute( self );
end;

procedure TTestCase.Teardown;
begin
    if Assigned( FOnTeardown ) then FOnTeardown( self );
end;

procedure TTestCase.Run;
begin
    if Assigned( TestResult ) then
        TestResult.RunTestCase( self )
    else
        RunDirectTest;
end;

procedure TTestCase.RunDirectTest;
begin
    Setup;
    try
      DoRunTest;
    finally
      Teardown;
    end;
end;


//==============================================================================
//  TTestSuite implementation
//==============================================================================

constructor TTestSuite.Create( AOwner : TComponent );
begin
    inherited Create( AOwner );
    FTests := TList.Create;
end;

destructor TTestSuite.Destroy;
begin
    FTests.Free;
    inherited Destroy;
end;

procedure TTestSuite.SetTestResult( newResult : TTestResult );
var
    oldResult   : TTestResult;
    iChild      : Integer;
    child       : TTest;
begin
    oldResult   := TestResult;

    inherited SetTestResult( newResult );

    for iChild  := 0 to FTests.Count - 1 do
    begin
        child   := TTest( FTests[ iChild ] );
        if child.TestResult = oldResult then
            child.TestResult := newResult;
    end;
end;

procedure TTestSuite.Notification( AComponent   : TComponent
                                 ; Operation    : TOperation );
begin
    if ( AComponent is TTest ) and ( Operation = opRemove ) then
        RemoveTest( TTest( AComponent ) );
end;

procedure TTestSuite.AddTest( child : TTest );
begin
    FTests.Add( child );
end;

procedure TTestSuite.RemoveTest( child : TTest );
begin
    FTests.Remove( child );
end;

procedure TTestSuite.Run;
var
    iTest   : Integer;

begin
    if Assigned( FOnBeginRun ) then FOnBeginRun( self );

    iTest := 0;
    while TestResult.OK and ( iTest < FTests.Count ) do
    begin
        TTest( FTests[ iTest ] ).Run;
        Inc( iTest );
    end;

    if Assigned( FOnEndRun ) then FOnEndRun( self );
end;


//==============================================================================
// Register components.
//==============================================================================

procedure Register;
begin
    RegisterComponents( 'DUnit', [ TTestResult, TTestCase, TTestSuite ] );
end;

end.
